/*
 * Search the registry for mime types and applications linked to them.
 * Try to query as much info as possible and create a .desktop application
 * entry.
 *
 * Copyright 2005 Stefan Dsinger
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <windows.h>
#include <stdio.h>

struct entry
{
	WCHAR *name;
	WCHAR *exp;
	WCHAR *mimetype;
	WCHAR *description;
	WCHAR *icon;
	WCHAR *apppath;
};

/*Store mime links and their applications*/
struct mimeapp
{
	struct mimeapp *prev;
	struct mimeapp *next;
	WCHAR *mimetype;
};

void usage()
{
	printf("No usage description yet\n");
}

BOOL get_mime_type(struct entry *app)
{
	HKEY hkey;
	BOOL ret=FALSE;
	long res;
	WCHAR data[256];
	DWORD datalen = 256;
	static const WCHAR CONTENT_TYPE[] = {'C','o','n','t','e','n','t',' ','T','y','p','e',0};
	
	if( res=RegOpenKeyExW(HKEY_CLASSES_ROOT, app->exp, 0, KEY_ALL_ACCESS, &hkey) != ERROR_SUCCESS )
		return FALSE;

	memset(data, 'a', 256*sizeof(WCHAR));
	res = RegQueryValueExW(hkey, CONTENT_TYPE, NULL, NULL, (LPBYTE) &data, &datalen);
	if(res == ERROR_SUCCESS)
	{
		app->mimetype = malloc((datalen+1)*sizeof(WCHAR));
		if(app->mimetype)
		{
			wcscpy(app->mimetype, data);

			if(RegQueryValueExW(hkey, NULL, NULL, NULL,(void *)&data, &datalen) == ERROR_SUCCESS)
			{
				app->name = malloc((datalen+1)*sizeof(WCHAR));
				if(app->name)
				{
					wcscpy(app->name, data);
//					printf("Got entry %s, mimetype %s\n", wine_dbgstr_w(app->name), wine_dbgstr_w(app->mimetype));
					ret=TRUE;
				}
			}
		}
	}

	RegCloseKey(hkey);
	return ret;
}

BOOL get_application(struct entry *app)
{
	HKEY hkey;
	HKEY iconkey;
	HKEY appkey;
	long res;
	WCHAR data[512];
	DWORD datalen = 512;
	BOOL ret=FALSE;
	WCHAR DEFAULTICON[] = {'D','e','f','a','u','l','t','I','c','o','n',0};
	WCHAR CMD[] = {'s','h','e','l','l','\\','o','p','e','n','\\','c','o','m','m','a','n','d',0};
	WCHAR DDE[] = {'s','h','e','l','l','\\','o','p','e','n','\\','d','d','e','e','x','e','c',0};

	if( res=RegOpenKeyExW(HKEY_CLASSES_ROOT, app->name, 0, KEY_ALL_ACCESS, &hkey) != ERROR_SUCCESS )
	{
		printf("Entry has no xxx\n");
		return FALSE;
	}
	res = RegQueryValueExW(hkey, NULL, NULL, NULL, (LPBYTE) &data, &datalen);
	if(res == ERROR_SUCCESS)
	{
		//printf("Got name %s\n", wine_dbgstr_w(data));
		app->description = malloc( (datalen+1) * sizeof(WCHAR));
		datalen=512;
		if(!app->description)
		{
			printf("Error: out of memory\n");
			return FALSE;
		}
		wcscpy(app->description, data);
		
		/*Get the app path.*/
		if(RegOpenKeyExW(hkey, CMD, 0, KEY_ALL_ACCESS, &appkey) == ERROR_SUCCESS )
		{
			res = RegQueryValueExW(appkey, NULL, NULL, NULL, (LPBYTE) &data, &datalen);
			if(res == ERROR_SUCCESS)
			{
//				printf("\tGot exec app %s\n", wine_dbgstr_w(data));
				app->apppath = malloc( (datalen+1) * sizeof(WCHAR));
				datalen=512;
				wcscpy(app->apppath, data);
			}
		}

		RegCloseKey(appkey);

		/*2nd chance: There might be a ddeexec key.*/
		if(!app->apppath &&  RegOpenKeyExW(hkey, DDE, 0, KEY_ALL_ACCESS, &appkey) == ERROR_SUCCESS )
		{
			res = RegQueryValueExW(appkey, NULL, NULL, NULL, (LPBYTE) &data, &datalen);
			if(res == ERROR_SUCCESS)
			{
//				printf("\tGot dde app %s\n", wine_dbgstr_w(data));
				app->apppath = malloc( (datalen+1) * sizeof(WCHAR));
				datalen=512;
				wcscpy(app->apppath, data);
			}
			else
				printf("Failed to get dde path 2 - might be a bug\n");
		}
		RegCloseKey(appkey);

		/*Get the icon. Require this, because I only accept fully qualified entries*/
		if(app->apppath && RegOpenKeyExW(hkey, DEFAULTICON, 0, KEY_ALL_ACCESS, &iconkey) == ERROR_SUCCESS )
		{
			res = RegQueryValueExW(iconkey, NULL, NULL, NULL, (LPBYTE) &data, &datalen);
			if(res == ERROR_SUCCESS)
			{
// 				printf("\tGot icon %s\n", wine_dbgstr_w(data));
				app->icon = malloc( (datalen+1) * sizeof(WCHAR));
				datalen=512;
				wcscpy(app->icon, data);
				ret = TRUE;
			}
			else
				printf("Failed to get Icon 2 - might be a bug\n");
		}
		else
// 			printf("Failed to get Icon 1 - Looks like there's no icon\n");
		RegCloseKey(iconkey);
	}
	
	RegCloseKey(hkey);
	return ret;
}

void write_desktop_file(struct entry *app, char *path)
{
	static int i=0;
	char *filename;
	FILE *outfile;
	char *tmp = malloc(512*sizeof(WCHAR));
	char *icon = NULL;
	WCHAR *iconpath;
	int iconindex;
	i++;	/*Next file number*/
	WCHAR *div = NULL;
	

	/*User info*/
	//printf("%d: Mime type %s:\n",i, wine_dbgstr_w(app->mimetype));
	//printf("\tExpansion %s:\n", wine_dbgstr_w(app->exp));
	//printf("\tRegistry name %s:\n", wine_dbgstr_w(app->name));
	//printf("\tFile Description %s:\n", wine_dbgstr_w(app->description));
	//printf("\tApplication%s:\n", wine_dbgstr_w(app->apppath));
	//printf("\tIcon: %s:\n", wine_dbgstr_w(app->icon));

	/*Write to the desktop file*/
	filename = malloc(12 + strlen(path) + 1);
	sprintf(filename, "%s%04d.desktop", path, i);
	//printf("\t%s:\n", filename);
	outfile = fopen(filename, "w");

	fprintf(outfile, "[Desktop Entry]\n");
	fprintf(outfile, "Version=1.0\n");
	fprintf(outfile, "Encoding=UTF-8\n");
	fprintf(outfile, "Type=Application\n");

	WideCharToMultiByte( CP_ACP, 0, app->name, wcslen(app->name)*2 + 1,
                             tmp, 512, NULL, NULL );
	fprintf(outfile, "Name=%s\n", tmp);

	WideCharToMultiByte( CP_ACP, 0, app->description, wcslen(app->description)*2 + 1,
                             tmp, 512, NULL, NULL );
	fprintf(outfile, "Comment=%s\n",tmp);

	fprintf(outfile, "Exec=winestart\n");

	/*Extract the icon*/
	if(app->icon)	
	{
		div = wcschr(app->icon, ',');
		if(div)
		{
			iconindex = wcstol(div + 1, NULL, 10);
			printf("\tIconindex %d\n", iconindex);
			*div = 0;
			icon =  (char *) extract_icon(app->icon, iconindex);
			if(icon)
			{
				printf("Got icon %s\n", icon);
			}
			else
				//printf("Failed to extract icon %s, %d\n", wine_dbgstr_w(app->icon), iconindex);
			WideCharToMultiByte( CP_ACP, 0, app->icon, wcslen(app->icon)*2 + 1,
        		                     tmp, 512, NULL, NULL );
			fprintf(outfile, "Icon=%s\n",tmp);
		}
	}
	WideCharToMultiByte( CP_ACP, 0, app->mimetype, wcslen(app->mimetype)*2 + 1,
                             tmp, 512, NULL, NULL );
	fprintf(outfile, "MimeType=%s\n", tmp);
	fprintf(outfile, "Terminal=false\n");
	fprintf(outfile, "StartupNotify=true\n");
	fclose(outfile);
	printf("\n\n");
}

BOOL check_doubles(struct entry *app, struct mimeapp *existing_items)
{
	struct mimeapp *elem = existing_items;
	BOOL exists = FALSE;
	while(elem->next)
	{
		if(elem->mimetype)
		{
			if(wcscmp(elem->mimetype, app->mimetype) != 0)
			{
				//printf("%s != %s\n",  wine_dbgstr_w(app->mimetype), wine_dbgstr_w(elem->mimetype));
			}
			else
			{
				//printf("Element %s exists\n", wine_dbgstr_w(app->mimetype));
				exists = TRUE;
				return FALSE;
			}
		}
		else
		{
			break;
		}
		if(elem->next)
			elem=elem->next;
	}
	if(!exists)
	{
		//printf("Adding element %s\n", wine_dbgstr_w(app->mimetype));
		if(!elem)
		{
			printf("autch\n");
			exit(3);
		}
		elem->next = malloc(sizeof(struct mimeapp));
		if(!elem->next)
		{
			printf("Error: out of memory\n");
			exit(3);
		}
		elem->next->prev = elem;
		elem->next->mimetype = NULL;
		elem->mimetype = malloc( (wcslen(app->mimetype) + 1) * sizeof(WCHAR));
		if(!elem->mimetype)
		{
			printf("Error: out of memory\n");
			exit(3);
		}
		wcscpy(elem->mimetype, app->mimetype);
		
	}
	return !exists;
}

int main(int argc, char *argv[])
{
	int argi;
	char *tpath = NULL;
	char *env = NULL;
	int len;
	int res;
	long subkeyindex = 0;
	WCHAR keyname[128];
	DWORD NameLen;
	HKEY Classes=NULL;
	NameLen=128;
	struct entry app;
	struct mimeapp existing_items;

	/*Where shall I write the application files to*/
	env = getenv("WINEAPPLNKPATH");
	if(!env)
	{
		printf("Error: You have to specify the target path in $WINEAPPLNKPATH\n");
		exit(3);
	}
	tpath = malloc(strlen(env) + 6);
	if(!tpath)
	{
		printf("Error: out of memory\n");
		exit(3);
	}
	strcpy(tpath, env);
	strcat(tpath, "/wine/");
	free(env);
	printf("Writing .desktop files to %s\n", tpath);

	/*Set up the list*/
	existing_items.next=NULL;
	existing_items.prev=NULL;
	existing_items.mimetype=NULL;

	/*Create the directory if it doesn't exist*/
	mkdir(tpath);

	/*Open the registry*/
	if( res=RegOpenKeyExW(HKEY_CLASSES_ROOT, NULL, 0, KEY_ALL_ACCESS, &Classes) != ERROR_SUCCESS )
	{
		printf("Error: Couldn't open HKEY_CLASSES_ROOT\n");
		exit(3);
	}

	while(RegEnumKeyExW(Classes, subkeyindex, keyname, &NameLen, NULL, NULL, NULL, NULL) == ERROR_SUCCESS)
	{
		app.name = NULL;
		app.exp = NULL;
		app.mimetype = NULL;
		app.description = NULL;
		app.icon = NULL;
		app.apppath = NULL;

		/*Is the key a file extension?*/
		if(keyname[0] == '.')
		{
			app.exp = malloc( ( NameLen+1 ) *sizeof(WCHAR) );
			if(!app.exp)
			{
				printf("Error: Out of memory\n");
				exit(3);
			}
			wcscpy(app.exp, keyname);

			/*Try to get all needed data*/
			if(get_mime_type(&app))
				if(get_application(&app))
					if(check_doubles(&app, &existing_items))
						write_desktop_file(&app, tpath);
		}
		
		/*Next key*/
		subkeyindex++;
		NameLen=128;
		if(app.name) free(app.name);
		if(app.exp) free(app.exp);
		if(app.mimetype) free(app.mimetype);
		if(app.description) free(app.description);
		if(app.icon) free(app.icon);
		if(app.apppath) free(app.apppath);
	}

	free(tpath);
	exit(0);
}
